//          Copyright Brian Schott (Hackerpilot) 2012.
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

module ctags;

import std.d.parser;
import std.d.lexer;
import std.d.ast;
import std.algorithm;
import std.range;
import std.stdio;
import std.array;
import std.conv;

/**
 * Prints CTAGS information to the given file.
 * Params:
 *     outpt = the file that CTAGS info is written to
 *     fileNames = tags will be generated from these files
 */
void printCtags(File output, string[] fileNames)
{
	string[] tags;
	LexerConfig config;
	StringCache cache = StringCache(StringCache.defaultBucketCount);
	foreach (fileName; fileNames)
	{
		File f = File(fileName);
		if (f.size == 0) continue;
		auto bytes = uninitializedArray!(ubyte[])(to!size_t(f.size));
		f.rawRead(bytes);
		auto tokens = getTokensForParser(bytes, config, &cache);
		Module m = parseModule(tokens.array, fileName, null, &doNothing);
		auto printer = new CTagsPrinter;
		printer.fileName = fileName;
		printer.visit(m);
		tags ~= printer.tagLines;
	}
	output.write("!_TAG_FILE_FORMAT\t2\n"
			~ "!_TAG_FILE_SORTED\t1\n"
			~ "!_TAG_FILE_AUTHOR\tBrian Schott\n"
			~ "!_TAG_PROGRAM_URL\thttps://github.com/Hackerpilot/Dscanner/\n");
	tags.sort().copy(output.lockingTextWriter);
}

private:

void doNothing(string, size_t, size_t, string, bool) {}

class CTagsPrinter : ASTVisitor
{
	override void visit(const ClassDeclaration dec)
	{
		tagLines ~= "%s\t%s\t%d;\"\tc\tline:%d%s\n".format(dec.name.text, fileName, dec.name.line, dec.name.line, context);
		const c = context;
		context = "\tclass:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const StructDeclaration dec)
	{
		tagLines ~= "%s\t%s\t%d;\"\ts\tline:%d%s\n".format(dec.name.text, fileName, dec.name.line, dec.name.line, context);
		const c = context;
		context = "\tstruct:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const InterfaceDeclaration dec)
	{
		tagLines ~= "%s\t%s\t%d;\"\ti\tline:%d%s\n".format(dec.name.text, fileName, dec.name.line, dec.name.line, context);
		const c = context;
		context = "\tclass:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const TemplateDeclaration dec)
	{
		tagLines ~= "%s\t%s\t%d;\"\tT\tline:%d%s\n".format(dec.name.text, fileName, dec.name.line, dec.name.line, context);
		const c = context;
		context = "\ttemplate:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const FunctionDeclaration dec)
	{
		tagLines ~= "%s\t%s\t%d;\"\tf\tarity:%d\tline:%d%s\n".format(dec.name.text, fileName,
			dec.name.line, dec.parameters.parameters.length, dec.name.line, context);
		const c = context;
		context = "\tfunction:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const Constructor dec)
	{
		tagLines ~= "this\t%s\t%d;\"\tf\tarity:%d\tline:%d%s\n".format(fileName,
			dec.line, dec.parameters.parameters.length, dec.line, context);
		const c = context;
		context = "\tfunction: this";
		dec.accept(this);
		context = c;
	}

	override void visit(const Destructor dec)
	{
		tagLines ~= "~this\t%s\t%d;\"\tf\tline:%d%s\n".format(fileName, dec.line,
			dec.line, context);
		const c = context;
		context = "\tfunction: this";
		dec.accept(this);
		context = c;
	}

	override void visit(const EnumDeclaration dec)
	{
		tagLines ~= "%s\t%s\t%d;\"\tg\tline:%d%s\n".format(dec.name.text, fileName,
			dec.name.line, dec.name.line, context);
		const c = context;
		context = "\tenum:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const UnionDeclaration dec)
	{
		if (dec.name == tok!"")
		{
			dec.accept(this);
			return;
		}
		tagLines ~= "%s\t%s\t%d;\"\tu\tline:%d%s\n".format(dec.name.text, fileName,
			dec.name.line, dec.name.line, context);
		const c = context;
		context = "\tunion:" ~ dec.name.text;
		dec.accept(this);
		context = c;
	}

	override void visit(const AnonymousEnumMember mem)
	{
		tagLines ~= "%s\t%s\t%d;\"\te\tline:%d%s\n".format(mem.name.text, fileName,
			mem.name.line, mem.name.line, context);
	}

	override void visit(const EnumMember mem)
	{
		tagLines ~= "%s\t%s\t%d;\"\te\tline:%d%s\n".format(mem.name.text, fileName,
			mem.name.line, mem.name.line, context);
	}

	override void visit(const VariableDeclaration dec)
	{
		foreach (d; dec.declarators)
		{
			tagLines ~= "%s\t%s\t%d;\"\tv\tline:%d%s\n".format(d.name.text, fileName,
				d.name.line, d.name.line, context);
		}
		dec.accept(this);
	}

	override void visit(const AutoDeclaration dec)
	{
		foreach (i; dec.identifiers)
		{
			tagLines ~= "%s\t%s\t%d;\"\tv\tline:%d%s\n".format(i.text, fileName,
				i.line, i.line, context);
		}
		dec.accept(this);
	}

	override void visit(const Invariant dec)
	{
		tagLines ~= "invariant\t%s\t%d;\"\tv\tline:%d%s\n".format(fileName, dec.line, dec.line, context);
	}

	alias visit = ASTVisitor.visit;

	string fileName;
	string[] tagLines;
	int suppressDepth;
	string context;
}

